/* program to decompose event signals for GRETINA / GRETA.

   Author:  D.C. Radford    Aug 2004
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "gdecomp.h"

double fitter(Event_Signal asig,  /* observed signals */
	      Event_Signal *bsig, /* fitted signals */
	      Interaction *ints, double *t0, int nints, int flag);
double eval_int_pos(Event_Signal asig,  /* observed signals */
		    Event_Signal *bsig, /* fitted signals */
		    Interaction *ints, double t0, int nints);
double coarse_grid_1(Event_Signal asig,  /* observed signals */
		     int seg, Interaction *ints,
		     double *chisq0, double min_e_fraction);
double coarse_grid_2(Event_Signal asig,  /* observed signals */
		     int seg1, int seg2, Interaction *ints,
		     double *chisq0, double min_e_fraction);
double refine_grid_1(Event_Signal asig,  /* observed signals */
		     double chisq, double chisq0, double min_e_fraction,
		     Interaction *ints);
double refine_grid_2(Event_Signal asig,  /* observed signals */
		     double chisq, double chisq0, double min_e_fraction,
		     Interaction *ints);

/* ================================================================ */

double decompose_1(Event_Signal asig,  /* observed signals */
		   Event_Signal *bsig, /* fitted signals */
		   int seg, Interaction *ints, double *t0, /* parameters */
		   int grid2, int fit1, int fit2, int fit3, int final_fit, /* control */
		   double min_e_fraction)
{
  /*
    performs the adaptive grid search algorithm for single-segment events,
    optionally followed by constrained least-squares

    returns chi-squared
    ints[] array is set to contain the resulting best-fit double-interaction
    grid2, fit1, fit2, fit3 and final_fit control what parts of the
    decomposition algorithm get run
  */

  double chisq0, chisq2, chisq_cg;
  int    i, best = 0;
  Interaction   f_ints[5][4];     /* test/fitted interactions */
  Event_Signal  f_bsig[5];        /* calculated/fitted event data */
  double        f_t0[5], chisq[5];

  *t0 = 0.0;
  chisq2 = 100.0;
  for (i=0; i<5; i++) {
    chisq[i] = 100.0;
    f_t0[i] = 0.0;
  }

  /* do the grid search over the course grid for the hit segment */
  chisq_cg = coarse_grid_1(asig, seg, f_ints[0], &chisq0, min_e_fraction);
  chisq[0] = eval_int_pos(asig, bsig, f_ints[0], *t0, 2);
  if (!quiet) {
    printf("** grid1: seg, chisq, pars: %2d, %7.4f, %7.4f", seg, chisq_cg, chisq[0]);
    for (i=0; i<2; i++) printf(", %6.3f %6.3f %6.3f %6.3f",
			       f_ints[0][i].r, f_ints[0][i].p,
			       f_ints[0][i].z, f_ints[0][i].e);
    printf("\n");
  }

  if (fit1) {
    /* do nonlinear least-squares fit
       starting from coarse grid point pair with lowest chisq */
    f_ints[1][0] = f_ints[0][0];
    f_ints[1][1] = f_ints[0][1];
    chisq[1] = fitter(asig, &f_bsig[1], f_ints[1], &f_t0[1], 2, 0);
    if (chisq[1] < chisq[best]) best = 1;
  }

  if (fit2 && f_ints[0][2].pos >= 0 && f_ints[0][3].pos >= 0) {
    /* try nonlinear least-squares fit
       starting from coarse grid point pair with second best chisq */
    f_ints[2][0] = f_ints[0][2];
    f_ints[2][1] = f_ints[0][3];
    chisq[2] = fitter(asig, &f_bsig[2], f_ints[2], &f_t0[2], 2, 0);
    if (chisq[2] < chisq[best]) best = 2;
  }

  if (grid2) {
    /* now do adaptive part of AGS,
       i.e. check neighboring sites on fine grid */
    chisq2 = refine_grid_1(asig, chisq_cg, chisq0, min_e_fraction, f_ints[0]);
    if (chisq2 < chisq_cg - 0.00000001) {
      chisq[0] = eval_int_pos(asig, bsig, f_ints[0], *t0, 2);
      printf("** fine_grid1: seg, chisq: %2d, %7.4f, %7.4f\n", seg, chisq2, chisq[0]);
      if (chisq[0] < chisq[best]) best = 0;
      if (fit3) {
	/* try nonlinear least-squares fit
	   starting from new fine grid point pair */
	f_ints[3][0] = f_ints[0][0];
	f_ints[3][1] = f_ints[0][1];
	chisq[3] = fitter(asig, &f_bsig[3], f_ints[3], &f_t0[3], 2, 0);
	if (chisq[3] < chisq[best]) best = 3;
      }
    }
  }

  /* decide which part of the algorithm gave the best chi-squared
     and, if necessary, finish up the nonlinear least-squares fit
     with more stringent convergence requirements */
  if (!quiet) printf(">>> g f1 f2 f3, best: %.4f %.4f %.4f %.4f, %.4f\n",
		     chisq[0], chisq[1], chisq[2], chisq[3], chisq[best]);

  ints[0] = f_ints[best][0];
  ints[1] = f_ints[best][1];
  *t0 = f_t0[best];
  if (best == 0) {
    if (!quiet) {  /* calculate bsig for matrix */
      chisq2 = eval_int_pos(asig, bsig, ints, *t0, 2);
      if (fabsf(chisq2 - chisq[best]) > 0.000001)
	printf("**** ACK!! best_chisq != eval_int_pos(); %f %f\n",
	       chisq[best], chisq2);
    }
  } else {
    if (final_fit) {
      chisq[best] = fitter(asig, bsig, ints, t0, 2, 1);
    } else {
      *bsig = f_bsig[best];
    }
  }

  return chisq[best];
} /* decompose_1 */


/* ================================================================ */

double decompose_2(Event_Signal asig,  /* observed signals*/
		   Event_Signal *bsig, /* fitted signals*/
		   int *seg, Interaction *ints, double *t0) /* parameters */
{
  /*
    performs the adaptive grid search algorithm for two-segment events,
    optionally followed by constrained least-squares

    returns chi-squared
    ints[] array is set to contain the resulting best-fit multi-interaction
  */

  double chisq0, chisq2, e2;
  int    i, j, fit1, fit2, fit3, fit4, grid2, shift_t0;
  int    pos, segsave, use_fit = 0, best = 0;
  Event_Signal  asig_save;
  Interaction   f_ints[10][4];     /* test/fitted interactions */
  Event_Signal  f_bsig[10];        /* calculated/fitted event data */
  double        f_t0[10], chisq[10];


  /* these next four lines control what parts of the decomposition
     algorithm get run */
  fit1  = 1;
  fit2  = 0;
  fit3  = 1;
  fit4  = 1;
  grid2 = 1;

  if (seg[0] > seg[1]) {
    segsave = seg[0];
    seg[0]  = seg[1];
    seg[1]  = segsave;
  }

  /* do the grid search for two interactions
     over the course grids for the two hit segments */
  for (i=0; i<10; i++) {
    chisq[i] = 100.0;
    f_t0[i] = 0.0;
  }

  /* CHECKME min. energy fract */
  chisq[0] = coarse_grid_2(asig, seg[0], seg[1], f_ints[0], &chisq0, 0.05);
  if (!quiet) {
    printf("** grid1a: seg0 seg1, chisq, pars: %2d %2d, %7.4f", seg[0], seg[1], chisq[0]);
    for (i=0; i<2; i++) printf(", %6.3f %6.3f %6.3f %6.3f",
			       f_ints[0][i].r, f_ints[0][i].p,
			       f_ints[0][i].z, f_ints[0][i].e);
    printf("\n");
  }

  if (fit1) {
    /* do nonlinear least-squares fit with 2 interactions,
       starting from coarse grid point pair with lowest chisq */
    f_ints[1][0] = f_ints[0][0];
    f_ints[1][1] = f_ints[0][1];
    chisq[1] = fitter(asig, &f_bsig[1], f_ints[1], &f_t0[1], 2, 0);
    if (chisq[1] < chisq[best]) best = 1;
  }

  if (grid2) {
    /* now do adaptive part of AGS,
       i.e. check neighboring sites on fine grid */
    /* CHECKME min. energy fract */
    chisq2 = refine_grid_2(asig, chisq[0], chisq0, 0.01, f_ints[0]);
    if (chisq2 < chisq[0]-0.00000001) {
      chisq[0] = chisq2;
      if (chisq[0] < chisq[best]) best = 0;
      if (fit2) {
 	/* try nonlinear least-squares fit
	   starting from new fine grid point pair */
	f_ints[2][0] = f_ints[0][0];
	f_ints[2][1] = f_ints[0][1];
	chisq[2] = fitter(asig, &f_bsig[2], f_ints[2], &f_t0[2], 2, 0);
	if (chisq[2] < chisq[best]) best = 2;
      }
    }
  }

  /* ----------------------------------------------
     now add a second interaction to the higher-energy segment
  */

  asig_save = asig;
  /* if necessary, switch the segments so that the highest energy is first */
  if (f_ints[best][0].e > f_ints[best][1].e) {
    f_ints[3][2] = f_ints[best][1];
  } else {
    segsave = seg[0];
    seg[0]  = seg[1];
    seg[1]  = segsave;
    f_ints[3][2] = f_ints[best][0];
  }
  f_t0[3] = f_t0[best];

  /* subtract the calculated signal for the lower-energy segment
     from the observed signal */
  if (best > 0) {
    /* best chisq is from a fit */
    use_fit = 1;
    eval_int_pos(asig, bsig, &f_ints[3][2], f_t0[3], 1);
    for (i=0; i<MEAS_SEGS; i++) {
      for (j=0; j<TIME_STEPS; j++) {
	asig.signal[i][j] -= bsig->signal[i][j];
      }
    }
    /* shift the remaining signal to bring t0 close to zero */
    shift_t0 = f_t0[best] - 1.0;
    if (shift_t0 > 0) {
      if (!quiet) printf("*----* shifting t0 by %d\n", shift_t0);
      for (i=0; i<MEAS_SEGS; i++) {
	for (j=0; j<TIME_STEPS-shift_t0; j++) {
	  asig.signal[i][j] = asig.signal[i][j+shift_t0];
	}
	for (j=TIME_STEPS-shift_t0; j<TIME_STEPS-1; j++) {
	  asig.signal[i][j] = asig.signal[i][TIME_STEPS-1];
	}
      }
    } else {
      shift_t0 = 0;
    }
  } else {
    /* best chisq is from AGS */
    pos = f_ints[3][2].pos;
    e2 = f_ints[3][2].e;
    for (i=0; i<MEAS_SEGS; i++) {
      for (j=0; j<TIME_STEPS; j++) {
	asig.signal[i][j] -= e2 * basis[pos].signal[i][j];
      }
    }
    shift_t0 = 0;
  }

  /* CHECKME min. energy fract */
  chisq[3] = decompose_1(asig, &f_bsig[3], seg[0], f_ints[3], &f_t0[3],
			 0, 1, 0, 0, 0, 0.01);
  if (!quiet) {
    printf("** grid1b: chisq, pars: %7.4f", chisq[3]);
    for (i=0; i<2; i++) printf(", %6.3f %6.3f %6.3f %6.3f",
			       f_ints[3][i].r, f_ints[3][i].p,
			       f_ints[3][i].z, f_ints[3][i].e);
    printf("\n");
  }
  f_t0[3] += (float) shift_t0;

  /* check chisq with full signal and 3 interactions */
  asig = asig_save;
  chisq2 = eval_int_pos(asig, bsig, f_ints[3], f_t0[3], 3);
  if (!quiet) {
    if (fabs(chisq[3]-chisq2) < 0.00001) {
      printf("***** chisq[3] == chisq2: %f %f\n", chisq[3], chisq2);
    } else {
      printf("***** chisq[3] != chisq2: %f %f\n", chisq[3], chisq2);
    }
  }
  chisq[3] = chisq2;
  if (chisq[3] < chisq[best]) best = 3;

  /* now we have three interactions,
     two in the higher-energy segment and one in the lower
     so perhaps we should try another least-squares fit with those three */

  if (fit3 || (fit4 && asig.seg_energy[seg[1]] < 80.0f)) {
    f_ints[4][0] = f_ints[3][0];
    f_ints[4][1] = f_ints[3][1];
    f_ints[4][2] = f_ints[3][2];
    f_t0[4] = f_t0[3];
    chisq[4] = fitter(asig, &f_bsig[4], f_ints[4], &f_t0[4], 3, 0);
    if (chisq[4] < chisq[best]) best = 4;
  }

  /* ----------------------------------------------
     now add a second interaction to the lower-energy segment
     but only if the segment energy is high enough
  */

  if (asig.seg_energy[seg[1]] >= 80.0f) {
    /* energy of lower segment is big enough for maybe 2 interactions */
    /* subtract the calculated signal for the higher-energy segment
       from the observed signal */
    for (i=0; i<MEAS_SEGS; i++) {
      for (j=0; j<TIME_STEPS; j++) {
	asig.signal[i][j] -= f_bsig[3].signal[i][j];
      }
    }
    shift_t0 = f_t0[3] - 1.0;
    if (shift_t0 > 0) {
      if (!quiet) printf("*----* shifting t0 by %d\n", shift_t0);
      for (i=0; i<MEAS_SEGS; i++) {
	for (j=0; j<TIME_STEPS-shift_t0; j++) {
	  asig.signal[i][j] = asig.signal[i][j+shift_t0];
	}
	for (j=TIME_STEPS-shift_t0; j<TIME_STEPS-1; j++) {
	  asig.signal[i][j] = asig.signal[i][TIME_STEPS-1];
	}
      }
    } else {
      shift_t0 = 0;
    }

    if (!quiet) printf("grid1c:\n");
    /* CHECKME min. energy fract */
    f_ints[5][0] = f_ints[3][0];
    f_ints[5][1] = f_ints[3][1];
    chisq[5] = decompose_1(asig, &f_bsig[5], seg[1], &f_ints[5][2], &f_t0[5],
			   0, 1, 0, 0, 0, 0.01);
    if (!quiet) {
      printf("** grid1c: chisq, pars: %7.4f", chisq[5]);
      for (i=0; i<4; i++) {
	printf(", %6.3f %6.3f %6.3f %6.3f",
	       f_ints[5][i].r, f_ints[5][i].p, f_ints[5][i].z, f_ints[5][i].e);
	if (i==1) printf("\n                               ");
      }
      printf("\n");
    }
    f_t0[5] += (float) shift_t0;
    /* let's not forget to restore the original observed signal */
    asig = asig_save;

    /* check chisq with full signal and 4 interactions */
    // f_t0[5] = (f_t0[3] + f_t0[5])/2.0;
    f_t0[5] = f_t0[3];
    chisq2 = eval_int_pos(asig, &f_bsig[5], f_ints[5], f_t0[5], 4);
    if (!quiet) {
      if (fabs(chisq[5]-chisq2) < 0.00001) {
	printf("***** chisq[5] == chisq2: %f %f\n", chisq[5], chisq2);
      } else {
	printf("***** chisq[5] != chisq2: %f %f\n", chisq[5], chisq2);
      }
    }
    chisq[5] = chisq2;
    if (chisq[5] < chisq[best]) best = 5;

    if (fit4) {
      f_ints[6][0] = f_ints[5][0];
      f_ints[6][1] = f_ints[5][1];
      f_ints[6][2] = f_ints[5][2];
      f_ints[6][3] = f_ints[5][3];
      f_t0[6] = f_t0[5];
      chisq[6] = fitter(asig, &f_bsig[6], f_ints[6], &f_t0[6], 4, 0);
      if (chisq[6] < chisq[best]) best = 6;
    }
  }

  if (!quiet) printf(">>> g f1 f2 f3 f4 f5 f6, best:\n"
		     " %.4f %.4f %.4f %.4f %.4f %.4f %.4f, %.4f\n",
		     chisq[0], chisq[1], chisq[2], chisq[3],
		     chisq[4], chisq[5], chisq[6], chisq[best]);

  if (best == 1 || best == 2) {
    chisq[best] = fitter(asig, bsig, f_ints[best], &f_t0[best], 2, 1);
  } else if (best == 4) {
    chisq[best] = fitter(asig, bsig, f_ints[best], &f_t0[best], 3, 1);
  } else if (best == 6) {
    chisq[best] = fitter(asig, bsig, f_ints[best], &f_t0[best], 4, 1);
  } else if (!quiet && best == 0) {
    chisq2 = eval_int_pos(asig, bsig, f_ints[best], f_t0[best], 2);
    if (fabs(chisq[best]-chisq2) >= 0.00001) {
      printf("ACK! best chisq != chisq2: %f %f\n", chisq[best], chisq2);
      chisq[best] = chisq2;
    }
  } else if (!quiet && best == 5) {
    *bsig = f_bsig[5];
  }

  *t0 = f_t0[best];
  if (best < 3) {
    ints[0] = f_ints[best][0];
    ints[1].e = 0.0;
    ints[1].seg = ints[0].seg;
    ints[2] = f_ints[best][1];
    ints[3].e = 0.0;
    ints[3].seg = ints[2].seg;
  } else if (best < 5) {
    ints[0] = f_ints[best][0];
    ints[1] = f_ints[best][1];
    ints[2] = f_ints[best][2];
    ints[3].e = 0.0;
    ints[3].seg = ints[2].seg;
  } else {
    ints[0] = f_ints[best][0];
    ints[1] = f_ints[best][1];
    ints[2] = f_ints[best][2];
    ints[3] = f_ints[best][3];
  }

  return chisq[best];
} /* decompose_2 */


/* ================================================================ */

double decompose_n(Event_Signal asig,  /* observed signals*/
		   Event_Signal *bsig, /* fitted signals*/
		   int *seg, Interaction *ints, double *t0,  /* parameters */
		   int nseg)
{
  /*
    performs the adaptive grid search algorithm for three-segment events,
    optionally followed by constrained least-squares

    returns chi-squared
    ints[] array is set to contain the resulting best-fit multi-interaction
  */

  double chisq2;
  int    i, j, jseg, jfit, jint, segsave, best, shift_t0;
  Event_Signal  asig_save;
  Interaction   f_ints[3*MAX_SEGS][3*MAX_SEGS];     /* test/fitted interactions */
  Event_Signal  f_bsig[3*MAX_SEGS];        /* calculated/fitted event data */
  double        f_t0[3*MAX_SEGS], chisq[3*MAX_SEGS];
  int           nints[3*MAX_SEGS];


  /* put segments in order of decreasing energy */
  for (i=1; i<nseg; i++) {
    for (j=i; j>0 && asig.seg_energy[seg[j-1]] < asig.seg_energy[seg[j]]; j--) {
      segsave  = seg[j-1];
      seg[j-1] = seg[j];
      seg[j]   = segsave;
    }
  }

  /* start off with a simple nseg-interaction fit, one in each hit segment */
  for (i=0; i<nseg; i++) {
    f_ints[0][i].seg = seg[i];
    f_ints[0][i].pos = -1;
    f_ints[0][i].r   = maxir[seg[i]]/2;
    f_ints[0][i].p   = maxip[seg[i]]/2;
    f_ints[0][i].z   = maxiz[seg[i]]/2;
    f_ints[0][i].e   = asig.seg_energy[seg[i]] / asig.total_energy;
  }
  nints[0] = nseg;
  chisq[0] = fitter(asig, &f_bsig[0], f_ints[0], &f_t0[0], nseg, 0);
  best = 0;
  asig_save = asig;
  jfit = 1;

  /* loop throught the hit segments, trying to add interactions one at a time
     if the net energy gets too small, then quit  */
  for (jseg=0; jseg<nseg && asig.seg_energy[seg[jseg]] > 80.0; jseg++) {

    /* ----------------------------------------------
       now add a second interaction to the jseg-th segment;
       first subtract the calculated signal for the other segments
       from the observed signal */

    jint = nints[best] - nseg + jseg;
    for (i=0; i<jint; i++) {
      f_ints[jfit][i] = f_ints[best][i];
    }
    for (i=jint + 1; i<nints[best]; i++) {
      f_ints[jfit][i-1] = f_ints[best][i];
    }
    eval_int_pos(asig, bsig, f_ints[jfit], f_t0[best], 2);
    for (i=0; i<MEAS_SEGS; i++) {
      for (j=0; j<TIME_STEPS; j++) {
	asig.signal[i][j] -= bsig->signal[i][j];
      }
    }

    /* shift the remaining signal to bring t0 close to zero */
    shift_t0 = f_t0[best] - 1.0;
    if (shift_t0 > 0) {
      if (!quiet) printf("*----* shifting t0 by %d\n", shift_t0);
      for (i=0; i<MEAS_SEGS; i++) {
	for (j=0; j<TIME_STEPS-shift_t0; j++) {
	  asig.signal[i][j] = asig.signal[i][j+shift_t0];
	}
	for (j=TIME_STEPS-shift_t0; j<TIME_STEPS-1; j++) {
	  asig.signal[i][j] = asig.signal[i][TIME_STEPS-1];
	}
      }
    } else {
      shift_t0 = 0;
    }

    /* do 2-interation grid search in segment seg[jseg] */
    /* CHECKME min. energy fract */
    chisq[jfit] = decompose_1(asig, &f_bsig[jfit], seg[jseg], &f_ints[jfit][jint], &f_t0[jfit],
			      0, 1, 0, 0, 0, 0.01);
    if (!quiet) {
      printf("** grid1[%d]: chisq, pars: %7.4f", jseg, chisq[1]);
      for (i=0; i<2; i++) printf(", %6.3f %6.3f %6.3f %6.3f",
				 f_ints[1][i].r, f_ints[1][i].p,
				 f_ints[1][i].z, f_ints[1][i].e);
      printf("\n");
    }
    asig = asig_save;
    for (i=jint + 1; i<nints[best]; i++) {
      f_ints[jfit][i+1] = f_ints[best][i];
    }
    f_t0[jfit] = (f_t0[jfit] + f_t0[best] + (double) shift_t0)/2.0;
    nints[jfit] = nints[best] + 1;
    chisq2 = eval_int_pos(asig, bsig, f_ints[jfit], f_t0[jfit], nints[jfit]);
    if (!quiet) {
      if (fabs(chisq[jfit]-chisq2) < 0.00001) {
	printf("***** chisq[%d] == chisq2: %f %f\n", jfit, chisq[jfit], chisq2);
      } else {
	printf("***** chisq[%d] != chisq2: %f %f\n", jfit, chisq[jfit], chisq2);
      }
    }
    chisq[jfit] = chisq2;
    if (chisq[jfit] < chisq[best]) best = jfit;

    /* do nonlinear least-squares fit with the extra interaction(s)*/
    for (i=0; i<nints[jfit]; i++) {
      f_ints[jfit+1][i] = f_ints[jfit][i];
    }
    f_t0[jfit+1] = f_t0[jfit];
    nints[jfit+1] = nints[jfit];
    jfit++;
    chisq[jfit] = fitter(asig, &f_bsig[jfit], f_ints[jfit], &f_t0[jfit], nints[jfit], 0);
    if (chisq[jfit] < chisq[best]) best = jfit;

    if (!quiet) {
      printf("**>>> jint, jfit, best, chisq[best], nints[best] : %d %d %d %f %d\n",
	     jint, jfit, best, chisq[best], nints[best]);
    }
    jfit++;
  }

  /* do a final fit with tighter convergence criteria */
  for (i=0; i<nints[best]; i++) {
    f_ints[jfit][i] = f_ints[best][i];
  }
  f_t0[jfit] = f_t0[best];
  nints[jfit] = nints[best];
  chisq[jfit] = fitter(asig, &f_bsig[jfit], f_ints[jfit], &f_t0[jfit], nints[jfit], 1);
  if (chisq[jfit] < chisq[best]) best = jfit;
  
  /* return best-fit values, 2 interactions per segment */
  *bsig = f_bsig[best];
  *t0 = f_t0[best];
  jint = 0;
  for (i=0; i<nseg; i++) {
    ints[2*i] = f_ints[best][jint++];
    if (jint < nints[best] &&
	f_ints[best][jint].seg == f_ints[best][jint-1].seg) {
      ints[2*i+1] = f_ints[best][jint++];
    } else {
      ints[2*i+1] = ints[2*i];
      ints[2*i+1].e = 0.0;
    }
  }

  return chisq[best];
} /* decompose_n */
