/* program to decompose event signals for GRETINA / GRETA.

   Author:  D.C. Radford    Aug 2004
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "gdecomp.h"

#define NEVENTS 1000     /* number of events to process, normally 10000 for timing tests */
#define ENERGY  662.0    /* keV; 662.0 or 1332.0 */  /* not really needed */

int      read_basis(void);
void     pclock(int print_flag);
int      grid_init(void);
double   decompose_1(Event_Signal asig,  /* observed signals */
		     Event_Signal *bsig, /* fitted signals */
		     int seg, Interaction *ints, double *t0, /* parameters */
		     int grid2, int fit1, int fit2, int fit3, int final_fit, /* control */
		     double min_e_fraction);
double   decompose_2(Event_Signal asig,  /* observed signals*/
		     Event_Signal *bsig, /* fitted signals*/
		     int *seg, Interaction *ints, double *t0); /* parameters */
double   decompose_n(Event_Signal asig,  /* observed signals*/
		     Event_Signal *bsig, /* fitted signals*/
		     int *seg, Interaction *ints, double *t0,  /* parameters */
		     int nseg);
long int lrintf(float x);
int      cyl_to_cart(int seg, double *pars, double*x, double *y, double*z, double *e);
int      nearest_place(int seg, double rin, double pin, double zin,
		       float *rdiff, float *pdiff, float *zdiff, int pos_out[8]);


int main(int argc, char **argv)
{
  double xfinal[MAX_PARS], yfinal[MAX_PARS], zfinal[MAX_PARS], efinal[MAX_PARS];
  double efrac, elo, dinp, chisq, chisq2;
  double x1, y1, z1, e1, x2, y2, z2, e2, esum;
  int    i, j, event_number, found, photopk, events_proc = 0;
  short  mat[4096];
  FILE   *event_file, *file_out = 0, *mat_file = 0;

  Event_Signal  asig;                   /* observed event data */
  Event_Signal  bsig;                   /* calculated/fitted event data */
  Interaction   ints[2*MAX_SEGS];       /* interactions */
  int           nseg;                   /* number of segments that fired in the event */
  int           seg[TOT_SEGS];          /* IDs of the segments that fired */
  double        t0;                     /* time-zero parameter */

  /* process possible command line options:
     -q = quiet (no debug output, used for timing tests)
  */
  quiet = 0;
  for (i=1; i<argc; i++) {
    if (!strcmp(argv[i], "-q")) quiet = 1;
  }

  /* open event data file for reading */
  if (!(event_file=fopen(EVENT_FILE, "r"))) {
    printf("ERROR -- Cannot open file %s for reading.\n", EVENT_FILE);
    return 1;
  }
  //if (!quiet) {
    /* open results output file for writing */
    if (!(file_out=fopen("output/test_gdecomp.out", "w+"))) {
      printf("ERROR -- Cannot open file output/test_gdecomp.out for writing.\n");
      return 1;
    }
    /* open matrix file for writing */
    if (!(mat_file=fopen("output/out.mat", "w+"))) {
      printf("ERROR -- Cannot open file output/out.mat for writing.\n");
      return 1;
    }
    //}

  /* read in basis signal data */
  if (read_basis()) return 1;

  /* initialize arrays for the grid-search algorithm */
  pclock(1);
  if (grid_init()) return 1;
  pclock(1);

  /* start of event decomposition loop over events */

  for (event_number = 0; event_number < NEVENTS; event_number++) {
    if (!quiet) printf("Event %d\n", event_number); fflush(stdout);

    /* read in an event */
    if (!fread(&asig, sizeof(Event_Signal), 1, event_file)) {
      printf("No more events in input file?\n"
	     "     %d events read, %d events processed.\n", event_number, events_proc);
      pclock(1);
      return 1;
    }
    if (asig.total_energy < 20.0) continue;

    /* determine which segments have net energy */
    nseg = 0;
    esum = 0.0;
    for (i=0; i<TOT_SEGS; i++) {
      seg[i] = 0;
      if (asig.seg_energy[i] > 30.0) {
	esum += asig.seg_energy[i];
	seg[nseg++] = i;
      }
    }
    if (nseg == 0) continue;

    if (!quiet) {
      printf("-----------\n"
	     " Event # %d,  total energy = %5.0fkeV\n"
	     "  %d segments hit: ",
	     event_number, asig.total_energy, nseg);
      for (i=0; i<nseg; i++) {
	printf(" %2d: %4.0fkeV ", seg[i], asig.seg_energy[seg[i]]);
      }
      printf("\n-----------\n");
    }

    if (fabs(esum - asig.total_energy) > 30.0) {
      if (!quiet) printf("***Error***  Sum energy (%.0f) != CC energy (%.0f)\n",
			 esum, asig.total_energy);
      continue;
    }

    if (nseg > MAX_SEGS) {
      if (!quiet) printf("ACKK! Too many segments! nseg =%d\n", nseg);
      continue;
    }

    /* clear out arrays */
    memset(&bsig, 0, sizeof(Event_Signal));
    memset(&ints, 0, 2*nseg*sizeof(Interaction));

    /* branch according to number of hit segments */
    if (nseg == 1) {
      //continue;
      chisq = decompose_1(asig, &bsig, seg[0], ints, &t0, 1, 1, 1, 1, 1, 0.1);

    } else if (nseg == 2) {
      //continue;
      chisq = decompose_2(asig, &bsig, seg, ints, &t0);

    } else {
      continue;
      //chisq = decompose_n(asig, &bsig, seg, ints, &t0, nseg);
    }
    if (chisq > 99.9) {
      printf("Bad Chi-squared = %f for event number %d\n", chisq, event_number);
      continue;
    }
    events_proc++;

    /* this was in the original lbl code -
       I'm not sure how they got it, or if it's really that useful */
    if (asig.total_energy < 140.0) {
      elo = 0.35;
    } else if (asig.total_energy < 400.0) {
      elo = 0.28;
    } else if (asig.total_energy < 900.0) {
      elo = 0.20;
    } else {
      elo = 0.15;
    }
    found = 0;  /* number of interactions found during decomposition */

    /* loop over the hit segments */
    for (i=0; i<2*nseg; i+=2) {
      cyl_to_cart(ints[i].seg, &ints[i].r, &x1, &y1, &z1, &e1);
      e1 *= asig.total_energy;
      if (ints[i+1].e == 0.0) {
	xfinal[found] = x1;
	yfinal[found] = y1;
	zfinal[found] = z1;
	efinal[found] = e1;
	found++;
	continue;
      }
      cyl_to_cart(ints[i+1].seg, &ints[i+1].r, &x2, &y2, &z2, &e2);
      e2 *= asig.total_energy;

      /* difference in position */
      dinp  = sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) + (z1-z2)*(z1-z2));
      /* energy fraction for first interaction */
      efrac = e1/(e1+e2);

      /* separation less than 2 mm, or one interaction is too weak */
      if (dinp < 2.0 || efrac > (1.0 - elo) || efrac < elo) {
	/* combine the interactions */
	xfinal[found] = x1*efrac + x2*(1.0-efrac);
	yfinal[found] = y1*efrac + y2*(1.0-efrac);
	zfinal[found] = z1*efrac + z2*(1.0-efrac);
	efinal[found] = (e1+e2);
	found++;
      } else {
	/* no reason to think that there's anything wrong;
	   take both interactions */
	xfinal[found] = x1;
	yfinal[found] = y1;
	zfinal[found] = z1;
	efinal[found] = e1;
	found++;
	xfinal[found] = x2;
	yfinal[found] = y2;
	zfinal[found] = z2;
	efinal[found] = e2;
	found++;
      }
    }

    if (!quiet) {
      /* was it a photopeak event? */  /* this part is not really needed */
      photopk = 0;
      if ((asig.total_energy > ENERGY - 3.5 &&
	   asig.total_energy < ENERGY + 3.5) ||
	  (ENERGY == 1332.0 &&
	   asig.total_energy > 1173.0 - 3.5 &&
	   asig.total_energy < 1173.0 + 3.5)) {
	photopk = 1;
      }

      /* write results to file_out */
      chisq2 = chisq / (float) (MEAS_SEGS * TIME_STEPS); // approx. degrees of freedom
      chisq2 /= (4.0/asig.total_energy)*(4.0/asig.total_energy);  // noise ~ 4 keV
      fprintf(file_out,
	      " %2d %2d %9.2f  %6.2f %24.6f %9.4f %4d\n",
	      found, photopk, asig.total_energy, t0, chisq,
	      chisq2, event_number);
      for (i=0; i<found; i++) {
	fprintf(file_out,
		" %7.2f %7.2f %7.2f %7.2f\n",
		xfinal[i], yfinal[i], zfinal[i], efinal[i]);
      }

      /* write signals to matrix file */
      if (event_number < 4096) {
	for (i=0; i<4096; i++) {
	  mat[i] = 0;
	}
	for (i=0; i<MEAS_SEGS; i++) {
	  for (j=0; j<TIME_STEPS; j++) {
	    mat[50*i + j] = lrintf(asig.signal[i][j] * 10000.0f);
	    mat[2000 + 50*i + j] = lrintf(bsig.signal[i][j] * 10000.0f);
	  }
	}
	fwrite(mat, 8192, 1, mat_file);
      }
    }

  }
  /* pclock(3); */
  pclock(1);
  free(basis);
  for (i=0; i<SSEG; i++) {
    free(ags1[i].da);
    free(ags1[i].db);
    free(ags2[i].da);
    free(ags2[i].db);
    for (j=i+1; j<SSEG; j++) {
      free(ags_dc[i][j]);
    }
  }
  return 0;
} /* main */

/* ================================================================ */

int cyl_to_cart(int seg, double *pars, double*x, double *y, double*z, double *e)
{
  float  rdiff, pdiff, zdiff, dist[8];
  int    j, ir, ip, iz, pos[8];

  *x = *y = *z = 0;

  /* get indices of eight nearest neighbor points */
  if (!nearest_place(seg, pars[0], pars[1], pars[2], &rdiff, &pdiff, &zdiff, pos)) {
    /* weighting factors */
    dist[0] = (1.0f-rdiff)*(1.0f-pdiff)*(1.0f-zdiff);
    dist[1] = (1.0f-rdiff)*(1.0f-pdiff)*      zdiff;
    dist[2] = (1.0f-rdiff)*      pdiff *(1.0f-zdiff);
    dist[3] = (1.0f-rdiff)*      pdiff *      zdiff;
    dist[4] =       rdiff *(1.0f-pdiff)*(1.0f-zdiff);
    dist[5] =       rdiff *(1.0f-pdiff)*      zdiff;
    dist[6] =       rdiff *      pdiff *(1.0f-zdiff);
    dist[7] =       rdiff *      pdiff *      zdiff;
    for (j=0; j<8; j++) {
      if (pos[j] >= 0) {
	*x += dist[j]*basis[pos[j]].x;
	*y += dist[j]*basis[pos[j]].y;
	*z += dist[j]*basis[pos[j]].z;
      }
    }
  } else {
    ir = pars[0] + 0.0001;
    ip = pars[1] + 0.0001;
    iz = pars[2] + 0.0001;
    if (ir<0) ir = 0;
    if (ip<0) ip = 0;
    if (iz<0) iz = 0;
    if (ir>maxir[seg]) ir = maxir[seg];
    if (ip>maxip[seg]) ip = maxip[seg];
    if (iz>maxiz[seg]) iz = maxiz[seg];
    j = grid_pos_lu[seg][ir][ip][iz];
    if (j < 0) return 1;
    *x = basis[j].x;
    *y = basis[j].y;
    *z = basis[j].z;
  }
  *e = pars[3];
  return 0;
} /* cyl_to_cart */

/* ================================================================ */

/* This is just a quick time profiling routine to see where the
   decomposition algorithm is spending all it's time.
   Author:  D.C. Radford    May 2003
*/

void pclock(int print_flag)
{
  static int s = 0, t1 = 0, t2;
  /* usage:
     pclock(0) on entry to timed code
     pclock(1) on exit from timed code, to print time interval
     pclock(2) on exit from timed code, to add time interval to sum
     pclock(3) to print out sum of time intervals

     Time increments seem to be in 1/100 seconds
     So for multiple short time intervals, statistical fluctuations may be important.
  */

  t2 = (int) clock();
  if (print_flag == 1) printf("*** clock, interval: %10d %10d\n", t2, t2 - t1);
  if (print_flag == 2) s += t2 - t1;
  if (print_flag == 3) {
    printf("*** clock sum: %10d\n", s);
    s = 0;
  }
  t1 = t2;
} /* pclock */
